package ch.psi.daq.queryrest.controller;

import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.stream.Collectors;

import javax.annotation.PostConstruct;
import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import javax.validation.Valid;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.http.MediaType;
import org.springframework.validation.Validator;
import org.springframework.web.bind.WebDataBinder;
import org.springframework.web.bind.annotation.InitBinder;
import org.springframework.web.bind.annotation.PathVariable;
import org.springframework.web.bind.annotation.RequestBody;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.ResponseBody;
import org.springframework.web.bind.annotation.RestController;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.google.common.collect.Lists;

import ch.psi.daq.common.ordering.Ordering;
import ch.psi.daq.domain.backend.Backend;
import ch.psi.daq.domain.config.DomainConfig;
import ch.psi.daq.domain.json.channels.info.ChannelInfos;
import ch.psi.daq.domain.query.ChannelNameRequest;
import ch.psi.daq.domain.query.DAQQueries;
import ch.psi.daq.domain.query.DAQQuery;
import ch.psi.daq.domain.query.channels.ChannelsRequest;
import ch.psi.daq.domain.query.channels.ChannelsResponse;
import ch.psi.daq.domain.query.operation.Aggregation;
import ch.psi.daq.domain.query.operation.AggregationType;
import ch.psi.daq.domain.query.operation.Compression;
import ch.psi.daq.domain.query.operation.QueryField;
import ch.psi.daq.domain.query.response.Response;
import ch.psi.daq.domain.query.response.ResponseFormat;
import ch.psi.daq.domain.query.transform.image.color.ColorModelType;
import ch.psi.daq.domain.query.transform.image.resize.ValueAggregation;
import ch.psi.daq.domain.request.validate.RequestProviderValidator;
import ch.psi.daq.queryrest.query.QueryManager;
import ch.psi.daq.queryrest.response.AbstractHTTPResponse;
import ch.psi.daq.queryrest.response.PolymorphicResponseMixIn;
import ch.psi.daq.queryrest.response.json.JSONHTTPResponse;

@RestController
public class QueryRestController {

   private static final Logger LOGGER = LoggerFactory.getLogger(QueryRestController.class);
   public static final String PARAMETERS_ROOT_PATH = "params";

   @Resource
   private ApplicationContext appContext;

   @Resource
   private Validator queryValidator;
   private Validator requestProviderValidator = new RequestProviderValidator();

   @Resource
   private QueryManager queryManager;

   @Resource
   private ObjectMapper objectMapper;

   private Response defaultResponse = new JSONHTTPResponse();

   @PostConstruct
   public void afterPropertiesSet() {}

   @InitBinder
   protected void initBinder(WebDataBinder binder) {
      if (binder.getTarget() != null) {
         if (requestProviderValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(requestProviderValidator);
         }
         if (queryValidator.supports(binder.getTarget().getClass())) {
            binder.addValidators(queryValidator);
         }
      }
   }

   @RequestMapping(value = DomainConfig.PATH_CHANNELS, method = {RequestMethod.GET, RequestMethod.POST},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ChannelsResponse> getChannels(@RequestBody(required = false) ChannelsRequest request)
         throws Throwable {
      return queryManager.getChannels(request);
   }

   /**
    * Query specific channel names, and return only those.
    * 
    * @param channelName part of (or full) channel name
    * @return Collection of channel names matching the specified input channel name
    * @throws Throwable in case something goes wrong
    */
   @RequestMapping(value = DomainConfig.PATH_CHANNELS + "/{channelName}", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<ChannelsResponse> getChannels(@PathVariable(value = "channelName") String channelName)
         throws Throwable {
      return getChannels(new ChannelsRequest(channelName));
   }

   /**
    * Queries for channels info
    * 
    * @param request the ChannelNameRequest
    * @return Collection of ChannelInfos
    * @throws Throwable in case something goes wrong
    */
   @RequestMapping(
         value = DomainConfig.PATH_CHANNELS_INFO,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody Collection<ChannelInfos> executeChannelInfoQuery(@RequestBody ChannelNameRequest request)
         throws Throwable {
      return queryManager.getChannelInfos(request);
   }

   /**
    * Accepts the properties for the {@link DAQQuery} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQuery} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param res the current {@link HttpServletResponse} instance
    * @throws Exception if reading the JSON string fails or if the subsequent call to
    *         {@link #executeQuery(DAQQuery, HttpServletResponse)} fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERY,
         method = RequestMethod.GET)
   public void executeQueryBodyAsString(@RequestParam String jsonBody, HttpServletResponse res) throws Exception {
      DAQQuery query = objectMapper.readValue(jsonBody, DAQQuery.class);
      executeQuery(query, res);
   }

   /**
    * Executes a single query.
    * 
    * @param query the {@link DAQQuery}
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @throws Exception thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERY,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQuery(@RequestBody @Valid DAQQuery query, HttpServletResponse res) throws Exception {
      executeQueries(new DAQQueries(query), res);
   }

   /**
    * Accepts the properties for the {@link DAQQueries} instance as stringified JSON query string.
    * 
    * @param jsonBody The {@link DAQQueries} properties sent as a JSON string, i.e. this is the
    *        stringified body of the POST request method
    * @param res the current {@link HttpServletResponse} instance
    * @throws Exception if reading the JSON string fails or if the subsequent call to
    *         {@link #executeQueries(DAQQueries, HttpServletResponse)} fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERIES,
         method = RequestMethod.GET)
   public void executeQueriesBodyAsString(@RequestParam String jsonBody, HttpServletResponse res) throws Exception {
      DAQQueries queries = objectMapper.readValue(jsonBody, DAQQueries.class);
      executeQueries(queries, res);
   }

   /**
    * Catch-all query method for getting data from the backend for both JSON and CSV requests.
    * <p>
    * The {@link DAQQueries} object will contain the concrete subclass based on the combination of
    * fields defined in the user's query. The AttributeBasedDeserializer decides which class to
    * deserialize the information into and has been configured (see
    * QueryRestConfig#afterPropertiesSet) accordingly.
    * 
    * @param queries the {@link DAQQueries}
    * @param res the {@link HttpServletResponse} instance associated with this request
    * @throws Exception thrown if writing to the output stream fails
    */
   @RequestMapping(
         value = DomainConfig.PATH_QUERIES,
         method = RequestMethod.POST,
         consumes = {MediaType.APPLICATION_JSON_VALUE})
   public void executeQueries(@RequestBody @Valid DAQQueries queries, HttpServletResponse res) throws Exception {
      try {
         LOGGER.debug("Executing queries '{}'", queries);

         Response response = queries.getResponseOrDefault(defaultResponse);
         if (response instanceof AbstractHTTPResponse) {
            ((AbstractHTTPResponse) response).respond(appContext, queries, res);
         } else {
            String message =
                  String.format(
                        "Expecting Response of type '%s' but received '%s'. Check JSON deserialization defined in '%s'",
                        AbstractHTTPResponse.class.getName(), response.getClass().getName(),
                        PolymorphicResponseMixIn.class.getName());
            LOGGER.error(message);
            throw new IllegalArgumentException(message);
         }


      } catch (Exception e) {
         LOGGER.error("Failed to execute query '{}'.", queries, e);
         throw e;
      }
   }

   /**
    * Returns the current list of {@link Ordering}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @Deprecated
   @RequestMapping(value = "ordering", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Ordering> getOrderingValuesDeprecated() {
      return Lists.newArrayList(Ordering.values());
   }
   
   /**
    * Returns the current list of {@link Ordering}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/ordering", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Ordering> getOrderingValues() {
      return Lists.newArrayList(Ordering.values());
   }

   /**
    * Returns the current list of {@link ResponseFormat}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @Deprecated
   @RequestMapping(value = "responseformat", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ResponseFormat> getResponseFormatValuesDeprecated() {
      return Lists.newArrayList(ResponseFormat.values());
   }
   
   /**
    * Returns the current list of {@link ResponseFormat}s available.
    * 
    * @return list of {@link Ordering}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/responseformat", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ResponseFormat> getResponseFormatValues() {
      return Lists.newArrayList(ResponseFormat.values());
   }

   /**
    * Returns the current list of {@link QueryField}s available.
    * 
    * @return list of {@link QueryField}s as String array
    */
   @Deprecated
   @RequestMapping(value = "queryfields", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<QueryField> getQueryFieldValuesDeprecated() {
      return Arrays.stream(QueryField.values())
            .filter(queryField -> queryField.isPublish())
            .collect(Collectors.toList());
   }
   
   /**
    * Returns the current list of {@link QueryField}s available.
    * 
    * @return list of {@link QueryField}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/queryfields", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<QueryField> getQueryFieldValues() {
      return Arrays.stream(QueryField.values())
            .filter(queryField -> queryField.isPublish())
            .collect(Collectors.toList());
   }

   /**
    * Returns the current list of {@link Aggregation}s available.
    * 
    * @return list of {@link Aggregation}s as String array
    */
   @Deprecated
   @RequestMapping(value = "aggregations", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Aggregation> getAggregationValuesDeprecated() {
      return Lists.newArrayList(Aggregation.values());
   }
   
   /**
    * Returns the current list of {@link Aggregation}s available.
    * 
    * @return list of {@link Aggregation}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/aggregations", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Aggregation> getAggregationValues() {
      return Lists.newArrayList(Aggregation.values());
   }

   /**
    * Returns the current list of {@link AggregationType}s available.
    * 
    * @return list of {@link AggregationType}s as String array
    */
   @Deprecated
   @RequestMapping(value = "aggregationtypes", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<AggregationType> getAggregationTypeValuesDeprecated() {
      return Lists.newArrayList(AggregationType.values());
   }
   
   /**
    * Returns the current list of {@link AggregationType}s available.
    * 
    * @return list of {@link AggregationType}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/aggregationtypes", method = {RequestMethod.GET},
         produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<AggregationType> getAggregationTypeValues() {
      return Lists.newArrayList(AggregationType.values());
   }

   /**
    * Returns the current list of {@link Backend} values.
    * 
    * @return list of {@link Backend}s as String array
    */
   @Deprecated
   @RequestMapping(value = "backends", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Backend> getBackendValuesDeprecated() {
      return Lists.newArrayList(Backend.values());
   }
   
   /**
    * Returns the current list of {@link Backend} values.
    * 
    * @return list of {@link Backend}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/backends", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Backend> getBackendValues() {
      return Lists.newArrayList(Backend.values());
   }

   /**
    * Returns the current list of {@link Compression}s available.
    * 
    * @return list of {@link Compression}s as String array
    */
   @Deprecated
   @RequestMapping(value = "compression", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Compression> getCompressionValuesDeprecated() {
      return Lists.newArrayList(Compression.values());
   }
   
   /**
    * Returns the current list of {@link Compression}s available.
    * 
    * @return list of {@link Compression}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/compression", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<Compression> getCompressionValues() {
      return Lists.newArrayList(Compression.values());
   }
   
   /**
    * Returns the current list of {@link ValueAggregation}s available.
    * 
    * @return list of {@link ValueAggregation}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/valueaggregations", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ValueAggregation> getValueAggregations() {
      return Lists.newArrayList(ValueAggregation.values());
   }
   
   /**
    * Returns the current list of {@link ValueAggregation}s available.
    * 
    * @return list of {@link ValueAggregation}s as String array
    */
   @RequestMapping(value = PARAMETERS_ROOT_PATH + "/colormodeltypes", method = {RequestMethod.GET}, produces = {MediaType.APPLICATION_JSON_VALUE})
   public @ResponseBody List<ColorModelType> getColorModelTypes() {
      return Lists.newArrayList(ColorModelType.values());
   }
}
